/*
 *  upkg: tool for manipulating Unreal Tournament packages.
 *  Copyright (C) 2009 Nick Bowler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>

#include <glib-object.h>

#include "upkg.h"
#include "module.h"
#include "uobject.h"
#include "exportable.h"

enum {
	MODE_INFO,
	MODE_EXPORT,
	MODE_MAX
};

int verbose = 0;

static const char *progname = "upkg";
static const char *sopts = "ivVH";
static const struct option lopts[] = {
	{ "info",     0, NULL, 'i' },
	{ "verbose",  0, NULL, 'v' },
	{ "version",  0, NULL, 'V' },
	{ "help",     0, NULL, 'H' },
	{ 0 }
};

void print_version(void)
{
	printf("%s\n", PACKAGE_STRING);
	puts("\
Copyright (C) 2009 Nick Bowler.\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law."
	);
}

void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options] package\n", progname);
}

void print_help(void)
{
	print_usage(stdout);
	puts("Detailed help coming soon.  Until then, I'll just list my options.");
	for (unsigned i = 0; lopts[i].name; i++) {
		const struct option *o = &lopts[i];
		printf("\t--%s", o->name);

		if (o->has_arg == 1) {
			printf("=val");
		} else if (o->has_arg == 2) {
			printf("[=val]");
		}

		printf(" (-%c)\n", o->val);
	}
}

void print_upkg_flags(const char *prefix, unsigned long flags)
{
	if (flags & UPKG_FLAG_ALLOW_DOWNLOAD)
		printf("%sAllowDownload\n", prefix);
	if (flags & UPKG_FLAG_CLIENT_OPTIONAL)
		printf("%sClientOptional\n", prefix);
	if (flags & UPKG_FLAG_SERVER_ONLY)
		printf("%sServerOnly\n", prefix);
	if (flags & UPKG_FLAG_BROKEN_LINKS)
		printf("%sBrokenLinks\n", prefix);
	if (flags & UPKG_FLAG_INSECURE)
		printf("%sInsecure\n", prefix);
	if (flags & UPKG_FLAG_REQUIRED)
		printf("%sRequired\n", prefix);
}

void print_guid(unsigned char guid[static 16])
{
	for (unsigned i = 0; i < 16; i++) {
		printf("%02hhx", guid[i]);
		if (i == 15)
			putchar('\n');
		else
			putchar(' ');
	}
}

void print_upkg_exports(struct upkg *pkg)
{
	for (unsigned i = 0; i < pkg->export_count; i++) {
		const char *name, *package, *group, *class;

		name  = upkg_export_name(pkg, i);
		class = upkg_export_class(pkg, i, &package, &group);

		if (!name || !class)
			continue;

		printf("%u - %s%s%s (%s.%s)\n", i+1,
			group ? group : "", group ? "." : "", name,
			package, class);
	}
}

int package_info(struct upkg *pkg)
{
	printf("Version: %u\n",  pkg->version);
	printf("License: %u\n",  pkg->license);
	printf("GUID: ");
	print_guid(pkg->guid);

	printf("Flags:   %lx\n", pkg->flags);
	if (verbose >= 1) {
		print_upkg_flags("\t", pkg->flags);
	}

	if (verbose >= 1) {
		printf("Names:   %lu\n", pkg->name_count);
		if (verbose >= 2) {
			for (unsigned long i = 0; i < pkg->name_count; i++) {
				printf("\t%s\n", upkg_get_name(pkg, i));
			}
		}
	}

	printf("Exports: %lu\n", pkg->export_count);
	if (verbose >= 1) print_upkg_exports(pkg);
	printf("Imports: %lu\n", pkg->import_count);

	return EXIT_SUCCESS;
}

int main(int argc, char **argv)
{
	struct upkg *pkg;
	unsigned mode = MODE_INFO;
	int opt, rc = EXIT_FAILURE;

	if (argc > 0) progname = argv[0];

	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch (opt) {
		case 'i':
			mode = MODE_INFO;
			break;
		case 'v':
			verbose++;
			break;
		case 'V':
			print_version();
			return EXIT_SUCCESS;
		case 'H':
			print_help();
			return EXIT_SUCCESS;
		default:
			print_usage(stderr);
			return EXIT_FAILURE;
		}
	}

	if (argv[optind] == NULL) {
		print_usage(stderr);
		return EXIT_FAILURE;
	}

	if (module_init() != 0)
		return EXIT_FAILURE;

	pkg = upkg_fopen(argv[optind]);
	if (!pkg) {
		fprintf(stderr, "failed to open package!\n");
		return EXIT_FAILURE;
	}

	switch (mode) {
	case MODE_INFO:
		rc = package_info(pkg);
		break;
	}

	upkg_close(pkg);
	module_exit();
	return rc;
}
